import 'dart:typed_data';

import 'package:api/authentication/authentication_providers/interfaces/authentication_validator.dart';
import 'package:api/authentication/authentication_providers/interfaces/authenticator.dart';
import 'package:api/exceptions/authentication_format_exception.dart';
import 'package:api/exceptions/not_authenticated_exception.dart';
import 'package:basic_utils/basic_utils.dart';
import 'dart:convert';

class RsaApplicationAuthenticationProvider implements 
  Authenticator<String, String>,
  AuthenticationValidator<String>
{
  static const String algorithmName = 'SHA-256/RSA';
  static const AuthenticationExceptionType authenticationExceptionType = AuthenticationExceptionType.applicationSecret;

  final RSAPrivateKey privateKey; // openssl genpkey -algorithm RSA -out private_key.pem -pkeyopt rsa_keygen_bits:2048
  final RSAPublicKey publicKey; // openssl rsa -pubout -in private_key.pem -out public_key.pem

  RsaApplicationAuthenticationProvider(String privateKey, String publicKey) :
    privateKey = CryptoUtils.rsaPrivateKeyFromPem(privateKey),
    publicKey = CryptoUtils.rsaPublicKeyFromPem(publicKey);

  @override
  Future<String> authenticate(String apiKey) {
    var signer = Signer(algorithmName)
      ..init(true, PrivateKeyParameter<RSAPrivateKey>(privateKey));

    var apiKeyBytes = utf8.encode(apiKey);
    var signature = signer.generateSignature(Uint8List.fromList(apiKeyBytes)) as RSASignature;

    return Future.value('$apiKey.${base64Encode(signature.bytes)}');
  }

  @override
  Future<void> validate(String signedApiKey) {
    var signedApiKeyParts = signedApiKey.split('.');
    if (signedApiKeyParts.length != 2) AuthenticationFormatException("RSA key must consist of two parts separated by a dot");
    
    var apiKeyBytes = utf8.encode(signedApiKeyParts[0]);

    RSASignature rsaSignature;
    try {
      rsaSignature = RSASignature(base64Decode(signedApiKeyParts[1])); 
    } on FormatException catch (_) {
      throw AuthenticationFormatException("RSA signature is not a valid RSA signature");
    } 

    var verifier = Signer(algorithmName)
      ..init(false, PublicKeyParameter<RSAPublicKey>(publicKey));

    var isVerified = verifier.verifySignature(
      Uint8List.fromList(apiKeyBytes),
      rsaSignature,
    );

    if(!isVerified) {
      throw NotAuthenticatedException(authenticationExceptionType);
    }

    return Future.value();
  }
}
