import 'package:angel3_orm/angel3_orm.dart';
import 'package:api/converters/angel3/user_converter.dart';
import 'package:api/databases/interfaces/user_database.dart';
import 'package:api/exceptions/not_in_database_exception.dart';
import 'package:api/models/databases/angel3/models.dart';
import 'package:api/models/domain/user.dart' as dm;
import 'package:uuid/uuid.dart';

class Angel3UserDatabase implements UserDatabase<String> {
  final QueryExecutor _executor;

  Angel3UserDatabase(this._executor);

  @override
  Future<dm.User<String>> create(dm.User<String> item) async {
    var query = UserQuery();
    query.values
      ..id = item.id
      ..username = item.username
      ..passwordHash = item.passwordHash
      ..name = item.name
      ..isEnabled = item.isEnabled
      ..createdAt = DateTime.now().toUtc()
      ..updatedAt = DateTime.now().toUtc();

    var insertedUser = await query.insert(_executor);
    return insertedUser.value.toDomainUser();
  }

  @override
  Future<dm.User<String>> delete(String id) async {
    var query = UserQuery();
    query.where!
      .id.equals(id);

    var user = await query.deleteOne(_executor);
    return user.value.toDomainUser();
  }

  @override
  Future<String> generateId() {
    return Future.value(Uuid().v4());
  }

  @override
  Future<dm.User<String>> getUserByUsernamePasswordHash(String username, String passwordHash) async {
    var query = UserQuery();
    query.where!
      ..username.equals(username)
      ..passwordHash.equals(passwordHash);

    var receivedUser = await query.getOne(_executor);
    if(!receivedUser.isPresent){
      throw NotInDatabaseException();
    }
    
    return receivedUser.value.toDomainUser();
  }

  @override
  Future<bool> isUsernameExisting(String username) async {
    var query = await _executor.query(
      usersTableName,
      '''
      SELECT COUNT(*)
      FROM $usersTableName
      WHERE $userUsernameColumnName = @username
      ''', // table/column is not possible to substitute
      {
        'username': username,
      },
    );

    return query.first.first > 0;
  }

  @override
  Future<dm.User<String>> read(String id) async {
    var query = UserQuery();
    query.where!
      .id.equals(id);

    var user = await query.getOne(_executor);
    return user.value.toDomainUser();
  }

  @override
  Stream<dm.User<String>> readAll() async* {
    var query = UserQuery();
    var users = await query.get(_executor);

    yield* Stream.fromIterable(users.map((u) => u.toDomainUser()));
  }

  @override
  Future<dm.User<String>> update(dm.User<String> item) async {
    var query = UserQuery();
    query.where!.id.equals(item.id);
    query.values
      ..id = item.id
      ..username = item.username
      ..passwordHash = item.passwordHash
      ..name = item.name
      ..isEnabled = item.isEnabled
      ..updatedAt = DateTime.now().toUtc();

    var user = await query.updateOne(_executor);
    return user.value.toDomainUser();
  }
}
