import 'dart:io';

import 'package:logging/logging.dart';
import 'package:path/path.dart';
import 'package:postgres/postgres.dart';

const String migrationsTableName = '_migrations';
const String versionColumnName = '"version"';
const String createdAtColumnName = 'created_at';

Future<void> migratePostgres (
  Endpoint endpoint,
  Logger logger,
  {
    SslMode sslMode = SslMode.verifyFull
  }
) async {
  var postgresMigrationConnection = await Connection.open(
    endpoint,
    settings: ConnectionSettings(
      sslMode: sslMode,
    )
  );

  logger.info('Setting up Postgres migration');
  var migrationsFile = File('lib/migrations/database/postgres/migration_scripts/migrations.sql');
  await _applyFile(postgresMigrationConnection, migrationsFile, logger);

  logger.info('Starting Postgres migration');
  var migrationsFilePattern = RegExp(r'\d{4}\.sql');
  var migrationsDirectory = Directory('lib/migrations/database/postgres/migration_scripts');

  var migrationState = await _getMigrationState(postgresMigrationConnection);

  if (await migrationsDirectory.exists()) {
    var directoryContent = migrationsDirectory
      .listSync()
      .whereType<File>()
      .where((f) => migrationsFilePattern.hasMatch(f.path))
      .toList()
      ..sort((f1, f2) => f1.path.compareTo(f2.path));

    if (migrationState != null) {
      logger.info('Postgres migration is currently at version $migrationState');
      directoryContent = directoryContent
        .where((f) => migrationState.compareTo(basenameWithoutExtension(f.path)) < 0) // f.path must be bigger
        .toList();
    }

    if (directoryContent.isNotEmpty) {
      for (var file in directoryContent) {
        await _applyFile(postgresMigrationConnection, file, logger);
      }

      var currentMigrationState = basenameWithoutExtension(directoryContent.last.path);
      logger.info('Updating migration version to $currentMigrationState');
      await _updateMigrationState(postgresMigrationConnection, currentMigrationState);
    } else {
      logger.info('Postgres is already up to date, migration skipped');
    }
  }
  logger.info('Successfully migrated Postgres');
}

Future<void> _applyFile(
  Connection connection,
  File file,
  Logger logger,
) async {
  var fileName = basename(file.path);
  logger.info('Applying file $fileName');
  
  var command = await file.readAsString();

  var statements = command.split(';').where((s) => s.isNotEmpty && s != '\n');
  for (var statement in statements) {
    logger.info(statement.trim());
    await connection.execute('$statement;');
  }
  logger.info('Successfully applied file $fileName');
}

Future<String?> _getMigrationState(
  Connection connection
) async {
  var query = '''SELECT $versionColumnName
FROM $migrationsTableName
ORDER BY $createdAtColumnName
DESC
LIMIT 1;''';

  var queryResult = await connection.execute(query);
  return queryResult.isEmpty ? null : queryResult.first.first as String?;
}

Future<void> _updateMigrationState(
  Connection connection,
  String migrationState,
) async {
  var query = Sql.named('''INSERT INTO $migrationsTableName ($versionColumnName, $createdAtColumnName)
VALUES (@version, now());''');

  await connection.execute(
    query,
    parameters: {
      'version': migrationState,
    }
  );
}
