import 'package:api/authentication/authentication_providers/interfaces/authenticator.dart';
import 'package:api/databases/interfaces/user_database.dart';
import 'package:api/exceptions/not_in_database_exception.dart';
import 'package:api/exceptions/username_exists_exception.dart';
import 'package:api/exceptions/wrong_username_password_exception.dart';
import 'package:api/factories/responses/interfaces/authentication_response_factory.dart';
import 'package:api/models/domain/user.dart';
import 'package:api/models/requests/authentication/user_login_request.dart';
import 'package:api/models/requests/authentication/user_register_request.dart';
import 'package:api/models/responses/authentication/interfaces/authentication_response.dart';
import 'package:api/services/interfaces/authentication_service.dart';
import 'package:api/utils/authentication/hash_password.dart';

class DefaultAuthenticationService<TUserId, TUserSecret>
  implements AuthenticationService<TUserSecret>
{
  final UserDatabase<TUserId> userDatabase;
  final Authenticator<User<TUserId>, TUserSecret> userAuthenticator;
  final AuthenticationResponseFactory<TUserSecret> authenticationResponseFactory;

  DefaultAuthenticationService(
    this.userDatabase,
    this.userAuthenticator,
    this.authenticationResponseFactory
  );

  @override
  Future<AuthenticationResponse<TUserSecret>> register(UserRegisterRequest request) async {
    var passwordHash = hashPassword(request.password);

    if (await userDatabase.isUsernameExisting(request.username)) {
      throw UsernameExistsException();
    }

    var user = User<TUserId>(
      await userDatabase.generateId(),
      request.username,
      passwordHash,
      request.name,
      false,
    );

    var createdUser = await userDatabase.create(user);
    var userSecret = await userAuthenticator.authenticate(createdUser);
    return authenticationResponseFactory.createInstance(userSecret);
  }

  @override
  Future<AuthenticationResponse<TUserSecret>> login(UserLoginRequest request) async {
    var passwordHash = hashPassword(request.password);

    User<TUserId> user;
    try {
      user = await userDatabase.getUserByUsernamePasswordHash(request.username, passwordHash);
    } on NotInDatabaseException catch (_) {
      throw WrongUsernamePasswordException();
    }

    var userSecret = await userAuthenticator.authenticate(user);
    return authenticationResponseFactory.createInstance(userSecret);
  }
}
