import 'package:api/authentication/authentication_providers/interfaces/authenticator.dart';
import 'package:api/databases/interfaces/user_database.dart';
import 'package:api/exceptions/username_exists_exception.dart';
import 'package:api/models/domain/authentication.dart';
import 'package:api/models/domain/user.dart';
import 'package:api/models/requests/user/user_update_request.dart';
import 'package:api/models/responses/user/user_update_response.dart';
import 'package:api/services/interfaces/user_service.dart';
import 'package:api/utils/authentication/hash_password.dart';

class DefaultUserService<TUserId, TUserSecret>
  implements UserService<TUserSecret>
{
  final UserDatabase<TUserId> userDatabase;
  final Authenticator<User<TUserId>, TUserSecret> userAuthenticator;

  DefaultUserService(
    this.userDatabase,
    this.userAuthenticator,
  );
  
  @override
  Future<UserUpdateResponse<TUserSecret>> update(UserUpdateRequest request) async {
    var currentUser = await userDatabase.read(request.userId);

    if (request.newUsername != null && await userDatabase.isUsernameExisting(request.newUsername!) && currentUser.username != request.newUsername) {
      throw UsernameExistsException();
    }

    var newPasswordHash = request.newPassword == null ? currentUser.passwordHash : hashPassword(request.newPassword!);
    var updatingUser = User(
      currentUser.id,
      request.newUsername ?? currentUser.username,
      request.newPassword == null ? currentUser.passwordHash : hashPassword(request.newPassword!),
      request.newName ?? currentUser.name,
      currentUser.isEnabled,
    );

    var updatedUser = await userDatabase.update(updatingUser);

    var hasUsernameChanged = request.newUsername != null && currentUser.username != request.newUsername;
    var hasPasswordChanged = request.newPassword != null && currentUser.passwordHash != newPasswordHash;
    var hasNameChanged = request.newName != null && currentUser.name != request.newName;

    if(hasUsernameChanged || hasPasswordChanged) {
      var userSecret = await userAuthenticator.authenticate(updatedUser);
      var authentication = Authentication(userSecret);
      return UserUpdateResponse(true, authentication);
    } else if(hasNameChanged) {
      return UserUpdateResponse(true, null);
    }

    return UserUpdateResponse(false, null);
  }
}
