import 'dart:io';

import 'package:api/converters/responses/exception/exception_response_converter.dart';
import 'package:api/exceptions/authentication_format_exception.dart';
import 'package:api/exceptions/not_authenticated_exception.dart';
import 'package:api/exceptions/not_in_database_exception.dart';
import 'package:api/exceptions/request_attribute_format_exception.dart';
import 'package:api/exceptions/request_attribute_wrong_type_exception.dart';
import 'package:api/exceptions/request_missing_attribute_exception.dart';
import 'package:api/exceptions/username_exists_exception.dart';
import 'package:api/exceptions/wrong_username_password_exception.dart';
import 'package:api/utils/responses/responses_util.dart';

// generalizes http backends
abstract class HttpUtils {
  static Future<TResponse> executeErrorHandled<TRequest, TResponse>(
    Future<TResponse> Function(TRequest request) requestHandler,
    TRequest request,
    TResponse Function(int status, String? body) exceptionToResponse,
  ) async {
    try {
      // Must return TResponse with status 200
      return await requestHandler(request);
    } 
    on NotAuthenticatedException catch (e) {
      return _handleException(HttpStatus.unauthorized, e, exceptionToResponse);
    } 
    on NotInDatabaseException catch (e) {
      return _handleException(HttpStatus.notFound, e, exceptionToResponse);
    }
    on RequestAttributeWrongTypeException catch (e) {
      return _handleException(HttpStatus.badRequest, e, exceptionToResponse);
    }
    on RequestMissingAttributeException catch (e) {
      return _handleException(HttpStatus.badRequest, e, exceptionToResponse);
    }
    on UsernameExistsException catch (e) {
      return _handleException(HttpStatus.conflict, e, exceptionToResponse);
    }
    on WrongUsernamePasswordException catch (e) {
      return _handleException(HttpStatus.notFound, e, exceptionToResponse);
    }
    on AuthenticationFormatException catch (e) {
      return _handleException(HttpStatus.badRequest, e, exceptionToResponse);
    }
    on RequestAttributeFormatException catch (e) {
      return _handleException(HttpStatus.badRequest, e, exceptionToResponse);
    }
    on Exception catch (_) {
      // don't reveal other errors as they might contain sensitive information
      return _handleException<Exception, TResponse>(HttpStatus.internalServerError, null, exceptionToResponse);
    }
  }

  static TResponse _handleException<TException extends Exception, TResponse>(
    int status,
    TException? e,
    TResponse Function(int status, String? body) exceptionToResponse,
  ) {
    String? responseBody;
    if (e != null) {
      var exceptionResponse = ResponsesUtil.createExceptionResponse(e);
      responseBody = exceptionResponse.toResponseBody();
    }

    return exceptionToResponse(status, responseBody);
  }
}
